// -*- Mode: vala; indent-tabs-mode: nil; tab-width: 4 -*-
/*-
 * Copyright (c) 2013-2014 Audience Developers (http://launchpad.net/pantheon-chat)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by: Tom Beckmann <tomjonabc@gmail.com>
 *              Corentin Noël <corentin@elementaryos.org>
 */

using Clutter;

enum PlayFlags {
    VIDEO         = (1 << 0),
    AUDIO         = (1 << 1),
    TEXT          = (1 << 2),
    VIS           = (1 << 3),
    SOFT_VOLUME   = (1 << 4),
    NATIVE_AUDIO  = (1 << 5),
    NATIVE_VIDEO  = (1 << 6),
    DOWNLOAD      = (1 << 7),
    BUFFERING     = (1 << 8),
    DEINTERLACE   = (1 << 9),
    SOFT_COLORBALANCE = (1 << 10)
}

/* use internal function while
 * https://bugzilla.gnome.org/show_bug.cgi?id=754567
 * isn't fixed
 */
[CCode (cprefix = "Gst", gir_namespace="GstVideo", gir_version="1.0", lower_case_cprefix="gst_video_", cheader_filename = "gst/video/video.h")]
namespace GstVideo {
    [CCode (cheader_filename = "gst/video/video.h")]
    public static extern bool calculate_display_ratio (out uint dar_n, out uint dar_d, uint video_width, uint video_height, uint video_par_n, uint video_par_d, uint display_par_n, uint display_par_d);
}

namespace Audience.Widgets {
    public class VideoPlayer : Actor {

        bool _playing;
        public bool playing {
            get {
                return _playing;
            }
            set {
                if (value == playing)
                    return;

                playbin.set_state (value ? Gst.State.PLAYING : Gst.State.PAUSED);
                _playing = value;
            }
        }

        public double progress {
            get {
                int64 length, prog;
                playbin.query_duration (Gst.Format.TIME, out length);
                playbin.query_position (Gst.Format.TIME, out prog);
                if (length == 0)
                    return 0;

                return prog / (double)length;
            }
            set {
                int64 length;
                playbin.query_duration (Gst.Format.TIME, out length);
                playbin.seek_simple (Gst.Format.TIME, Gst.SeekFlags.FLUSH | Gst.SeekFlags.ACCURATE, (int64)(double.max (value, 0.0) * length));
            }
        }

        public double volume {
            get {
                return playbin.volume;
            }
            set {
                playbin.volume = value;
            }
        }

        public string uri {
            owned get {
                return playbin.current_uri;
            }
            set {
                if (value == (string)playbin.uri)
                    return;

                playbin.get_bus ().set_flushing (true);
                playing = false;
                playbin.set_state (Gst.State.READY);
                playbin.suburi = null;
                subtitle_uri = null;
                playbin.get_bus ().set_flushing (false);
                playbin.uri = value;
                volume = 1.0;

                relayout ();
                playing = false;

                // this has to be done after setting uri playbin, if not codec checking will fail
                try {
                    var info = new Gst.PbUtils.Discoverer (10 * Gst.SECOND).discover_uri (value);
                    var video = info.get_video_streams ();
                    if (video != null && video.data != null) {
                        var video_info = (Gst.PbUtils.DiscovererVideoInfo)video.data;
                        uint w, h;
                        get_media_size (video_info, out w, out h);
                        video_width = w;
                        video_height = h;
                    }
                } catch (Error e) {
                    warning (e.message);
                    return;
                }
            }
        }

        public int current_audio {
            get {
                return playbin.current_audio;
            }
            set {
                playbin.current_audio = value;
            }
        }

        string? subtitle_uri = null;

        // currently used text stream. Set to -1 to disable subtitles
        public int current_text {
            get {
                return playbin.current_text;
            }
            set {
                if (value == current_text)
                    return;

                int flags;
                playbin.get ("flags", out flags);

                var disable = value < 0;
                if (disable)
                    playbin.current_text = -1;

                check_text_layer (!disable);
                if (!disable) {
                    playbin.current_text = value;
                }
            }
        }

        private string _subtitle_font;
        public string subtitle_font {
            get {
                return _subtitle_font;
            }
            set {
                _subtitle_font = value;
                playbin.set ("subtitle-font-desc", _subtitle_font);
            }
        }

        public dynamic Gst.Element playbin;
        Clutter.Texture video;

        public uint video_width { get; private set; }
        public uint video_height { get; private set; }

        public signal void ended ();
        public signal void toggle_side_pane (bool show);
        public signal void text_tags_changed ();
        public signal void audio_tags_changed ();
        public signal void error ();
        public signal void plugin_install_done ();
        public signal void progression_changed (double current_time, double total_time);
        public signal void external_subtitle_changed (string? uri);

        public VideoPlayer () {
            video = new Clutter.Texture ();
            playbin = Gst.ElementFactory.make ("playbin", "playbin");
            build ();

            playbin.text_tags_changed.connect ((el) => {
                var structure = new Gst.Structure.empty ("tags-changed");
                structure.set_value ("type", "text");
                el.post_message (new Gst.Message.application (el, (owned) structure));
            });

            playbin.audio_tags_changed.connect ((el) => {
                var structure = new Gst.Structure.empty ("tags-changed");
                structure.set_value ("type", "audio");
                el.post_message (new Gst.Message.application (el, (owned) structure));
            });

            Timeout.add (100, () => {
                int64 length, prog;
                playbin.query_position (Gst.Format.TIME, out prog);
                playbin.query_duration (Gst.Format.TIME, out length);
                if (length == 0)
                    return true;

                progression_changed ((double)prog, (double)length);
                return true;
            });

            playbin.get_bus ().add_signal_watch ();
            playbin.get_bus ().message.connect (watch);
        }

        ~VideoPlayer () {
            playbin.set_state (Gst.State.NULL);
            playbin.get_bus ().message.disconnect (watch);
            message ("video player destroyed");
        }

        public void build (){

            if (video.get_parent ()!=null)
                remove_child (video);
            dynamic Gst.Element video_sink = Gst.ElementFactory.make ("cluttersink", "source");
            video_sink.texture = video;
            playbin.video_sink = video_sink;
            add_child(video);
            update_subtitle_font ();
            settings.changed.connect (() => {
                update_subtitle_font ();
            });
        }

        void watch (Gst.Message msg) {

            switch (msg.type) {
                case Gst.MessageType.APPLICATION:
                    if (msg.get_structure ().get_name () == "tags-changed") {
                        if (msg.get_structure ().get_string ("type") == "text")
                            text_tags_changed ();
                        else
                            audio_tags_changed ();
                    }
                    break;
                case Gst.MessageType.ERROR:
                    GLib.Error e; string detail;
                    msg.parse_error (out e, out detail);
                    playbin.set_state (Gst.State.NULL);

                    warning (detail);

                    show_error (e.message);
                    break;
                case Gst.MessageType.EOS:
                    Idle.add (()=>{
                            ended ();
                            return false;
                            });
                    break;
                case Gst.MessageType.ELEMENT:
                    if (msg.get_structure () == null)
                        break;

                    if (Gst.PbUtils.is_missing_plugin_message (msg)) {
                        error ();
                        playbin.set_state (Gst.State.NULL);

                        handle_missing_plugin (msg);
                    /*TODO } else { //may be navigation command
                        var nav_msg = Gst.Navigation.message_get_type (msg);

                        if (nav_msg == Gst.NavigationMessageType.COMMANDS_CHANGED) {
                            var q = Gst.Navigation.query_new_commands ();
                            pipeline.query (q);

                            uint n;
                            gst_navigation_query_parse_commands_length (q, out n);
                            for (var i=0;i<n;i++) {
                                Gst.NavigationCommand cmd;
                                gst_navigation_query_parse_commands_nth (q, 0, out cmd);
                                debug ("Got command: %i", (int)cmd);
                            }
                        }*/
                    }
                    break;
                default:
                    break;
            }
        }

        public void set_subtitle_uri (string? uri) {
            subtitle_uri = uri;
            if (!check_text_layer (subtitle_uri != null)) {
                apply_subtitles ();
                external_subtitle_changed (uri);
            };
        }

        // checks whether text layer has to be enabled
        // returns if apply_subtitles has been called
        bool check_text_layer (bool enable) {
            int flags;
            playbin.get ("flags", out flags);

            if (!enable && (flags & PlayFlags.TEXT) > 0) {
                flags &= ~PlayFlags.TEXT;
                playbin.set ("flags", flags);
            } else if (enable && (flags & PlayFlags.TEXT) < 1) {
                flags |= PlayFlags.TEXT;
                playbin.set ("flags", flags);
                apply_subtitles ();
                return true;
            }

            return false;
        }

        // ported from totem bvw widget set_subtitle_uri
        void apply_subtitles () {
            int64 time;
            playbin.query_position (Gst.Format.TIME, out time);

            playbin.get_state (null, null, Gst.CLOCK_TIME_NONE);

            Gst.State current;
            playbin.get_state (out current, null, Gst.CLOCK_TIME_NONE);
            if (current > Gst.State.READY) {
                playbin.set_state (Gst.State.READY);
                playbin.get_state (null, null, Gst.CLOCK_TIME_NONE);
            }

            playbin.suburi = subtitle_uri;
            if (current > Gst.State.READY) {
                playbin.set_state (current);
                playbin.get_state (null, null, Gst.CLOCK_TIME_NONE);
            }

            playbin.set_state (Gst.State.PAUSED);
            playbin.seek (1.0, Gst.Format.TIME,
                    Gst.SeekFlags.FLUSH | Gst.SeekFlags.ACCURATE,
                    Gst.SeekType.SET, time,
                    Gst.SeekType.NONE, (int64)Gst.CLOCK_TIME_NONE);

            if (current > Gst.State.READY) {
                playbin.set_state (current);
                playbin.get_state (null, null, Gst.CLOCK_TIME_NONE);
            }
        }

        public bool relayout () {
            if (video_width < 1 || video_height < 1 || uri == null)
                return false;

            var stage = get_stage ();
            var aspect = stage.width / video_width < stage.height / video_height ?
                stage.width / video_width : stage.height / video_height;
            video.width  = video_width * aspect;
            video.height = video_height * aspect;
            video.x = (stage.width  - video.width)  / 2;
            video.y = (stage.height - video.height) / 2;

            return true;
        }

        void show_error (string? message=null) {
            var dlg  = new Gtk.Dialog.with_buttons (_("Error"), null, Gtk.DialogFlags.MODAL, _("_OK"), Gtk.ResponseType.OK);
            var grid = new Gtk.Grid ();
            var err  = new Gtk.Image.from_icon_name ("dialog-error", Gtk.IconSize.DIALOG);
            err.margin_right = 12;

            var err_label = new Gtk.Label ("");
            err_label.set_markup ("<b>%s</b>".printf (_("Oops! Audience can't play this file!")));

            grid.margin = 12;
            grid.attach (err, 0, 0, 1, 1);
            grid.attach (err_label, 1, 0, 1, 1);
            if (message != null)
                grid.attach (new Gtk.Label (message), 1, 1, 1, 2);

            error ();
            ((Gtk.Box)dlg.get_content_area ()).add (grid);
            dlg.show_all ();
            dlg.run ();
            dlg.destroy ();
        }

        void handle_missing_plugin (Gst.Message msg) {
            var detail = Gst.PbUtils.missing_plugin_message_get_description (msg);
            var dlg = new Gtk.Dialog.with_buttons ("Missing plugin", null, Gtk.DialogFlags.MODAL);
            var grid = new Gtk.Grid ();
            var err  = new Gtk.Image.from_icon_name ("dialog-error", Gtk.IconSize.DIALOG);
            var phrase = new Gtk.Label (_("Some media files need extra software to be played. Audience can install this software automatically."));

            err.margin_right = 12;

            var err_label = new Gtk.Label ("");
            err_label.set_markup ("<b>%s</b>".printf (_("Audience needs %s to play this file.").printf (detail)));

            grid.margin = 12;
            grid.attach (err, 0, 0, 1, 1);
            grid.attach (err_label, 1, 0, 1, 1);
            grid.attach (phrase, 1, 1, 1, 2);

            dlg.add_button (_("Don't install"), 1);
            dlg.add_button (_("Install")+" "+detail, 0);

            (dlg.get_content_area () as Gtk.Container).add (grid);
            dlg.show_all ();
            if (dlg.run () == 0) {
                var installer = Gst.PbUtils.missing_plugin_message_get_installer_detail (msg);
                var context = new Gst.PbUtils.InstallPluginsContext ();
                Gst.PbUtils.install_plugins_async ({installer}, context, () => { //finished
                    debug ("Finished plugin install");
                    Gst.update_registry ();
                    plugin_install_done ();
                    playing = true;
                });
            }

            dlg.destroy ();
        }

        public void seek_jump_seconds (int seconds) {
            int64 position;
            playbin.query_position (Gst.Format.TIME, out position);

            var gst_seconds = 1000000000 * (int64)seconds;
            var new_position = position + gst_seconds;

            if (new_position < 0) {
                playbin.seek_simple (Gst.Format.TIME, Gst.SeekFlags.FLUSH | Gst.SeekFlags.ACCURATE, int64.max (new_position, 1));
                return;
            }

            playbin.seek_simple (Gst.Format.TIME, Gst.SeekFlags.FLUSH | Gst.SeekFlags.ACCURATE, new_position);
        }



        private void get_media_size (Gst.PbUtils.DiscovererVideoInfo video_info, out uint video_width, out uint video_height) {
            var video_w = video_info.get_width ();
            var video_h = video_info.get_height ();
            var video_par_n = video_info.get_par_num ();
            var video_par_d = video_info.get_par_denom ();
            if (video_par_n == 0 || video_par_d == 0) {
                debug ("width and/or height 0, assumming 1/1");
                video_par_n = 1;
                video_par_d = 1;
            }
            debug ("video PAR %u/%u", video_par_n, video_par_d);

            uint num;
            uint den;
            if (!GstVideo.calculate_display_ratio (out num, out den, video_w, video_h, video_par_n, video_par_d, 1, 1)) {
                warning ("overflow calculating display ratio");
                num = 1;
                den = 1;
            }
            debug ("calculated scale ratio %u / %u for %ux%u", num, den, video_w, video_h);

            if (video_w % num == 0) {
                debug ("keeping video width");
                video_width = video_w;
                video_height = (uint) Gst.Util.uint64_scale (video_w, den, num);
            } else {
                debug ("keeping video height");
                video_width = (uint) Gst.Util.uint64_scale (video_h, num, den);
                video_height = video_h;
            }
            debug ("scaling to %ux%u", video_width, video_height);
        }

        void update_subtitle_font () {
            debug ("Updating subtitle font");
            var font = settings.subtitle_font;

            if (font == "") {
                var gnome_settings = new GLib.Settings ("org.gnome.desktop.interface");
                font = gnome_settings.get_string ("font-name");
            }
            subtitle_font = font;
        }
    }
}
